#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <float.h>
#include <vector>

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include "mesh.h"

bool init();
void cleanup();
void disp();
void idle();
void draw_text(float r, float g, float b, const char *fmt, ...);
void reshape(int x, int y);
void keyb(unsigned char key, int x, int y);
void mouse(int bn, int state, int x, int y);
void motion(int x, int y);

float cam_theta = 0, cam_phi = 25.0, cam_dist = 10.0;
int win_width, win_height;

std::vector<Mesh*> meshes;
int selected = -1;

bool show_bspheres = false;

int main(int argc, char **argv)
{
	glutInit(&argc, argv);
	glutInitWindowSize(800, 600);
	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
	glutCreateWindow("foo");

	glutDisplayFunc(disp);
	glutIdleFunc(idle);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyb);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);

	if(!init()) {
		return 1;
	}
	atexit(cleanup);

	glutMainLoop();
	return 0;
}


bool init()
{
	win_width = glutGet(GLUT_WINDOW_WIDTH);
	win_height = glutGet(GLUT_WINDOW_HEIGHT);

	glClearColor(0.1, 0.1, 0.1, 1);

	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	Mesh *mesh = new Mesh;
	if(!mesh->load("mesh1.obj")) {
		return false;
	}
	mesh->set_position(-2, 0, 0);
	meshes.push_back(mesh);

	mesh = new Mesh;
	if(!mesh->load("mesh2.obj")) {
		return false;
	}
	mesh->set_position(2, 0, 0);
	meshes.push_back(mesh);

	return true;
}

void cleanup()
{
	for(int i=0; i<(int)meshes.size(); i++) {
		delete meshes[i];
	}
}

void disp()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	/* set the light in view space, before the view transform, to position it
	 * in relation to the camera instead of the world
	 */
	float lpos[] = {-30, 40, 50, 1};
	glLightfv(GL_LIGHT0, GL_POSITION, lpos);

	glTranslatef(0, 0, -cam_dist);
	glRotatef(cam_phi, 1, 0, 0);
	glRotatef(cam_theta, 0, 1, 0);


	for(int i=0; i<(int)meshes.size(); i++) {
		if(selected == i) {
			float diffuse[] = {1.0, 0.3, 0.2, 1.0};
			glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, diffuse);
		} else {
			float diffuse[] = {0.2, 0.3, 1.0, 1.0};
			glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, diffuse);
		}

		float specular[] = {1, 1, 1, 1};
		glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
		glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 60.0);

		if(show_bspheres) {
			meshes[i]->draw_bsph();
		} else {
			meshes[i]->draw();
		}
	}

	// draw a floor
	float floor_color[] = {0.3, 0.5, 0.35, 1};
	float black[] = {0, 0, 0, 1};
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, floor_color);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, black);

	glBegin(GL_QUADS);
	glNormal3f(0, 1, 0);
	glVertex3f(-10, 0, 10);
	glVertex3f(10, 0, 10);
	glVertex3f(10, 0, -10);
	glVertex3f(-10, 0, -10);
	glEnd();

	glutSwapBuffers();
	assert(glGetError() == GL_NO_ERROR);
}

void idle()
{
	glutPostRedisplay();
}

void reshape(int x, int y)
{
	win_width = x;
	win_height = y;

	glViewport(0, 0, x, y);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(50.0, (float)x / (float)y, 0.5, 500);
}

void keyb(unsigned char key, int x, int y)
{
	switch(key) {
	case 27:
	case 'q':
	case 'Q':
		exit(0);

	case 'b':
		show_bspheres = !show_bspheres;
		glutPostRedisplay();
	}
}

static int bnstate[16];
static int prev_x, prev_y;

void mouse(int bn, int state, int x, int y)
{
	int idx = bn - GLUT_LEFT_BUTTON;
	bnstate[idx] = state == GLUT_DOWN;

	prev_x = x;
	prev_y = y;
}


void motion(int x, int y)
{
	int dx = x - prev_x;
	int dy = y - prev_y;
	prev_x = x;
	prev_y = y;

	if(bnstate[0]) {
		cam_theta += dx * 0.5;
		cam_phi += dy * 0.5;
		if(cam_phi < -90) cam_phi = -90;
		if(cam_phi > 90) cam_phi = 90;
		glutPostRedisplay();
	}

	if(bnstate[2]) {
		cam_dist += dy * 0.1;
		if(cam_dist < 0.0) {
			cam_dist = 0.0;
		}
		glutPostRedisplay();
	}
}
